/* GTK Napster-0.201
 * Copyright (C) 2000 Unified (xilliator@hotmail.com).
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <sys/time.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>

#include "common.h"
#include "dlsess.h"

void cancel_download(GtkWidget *widget, gpointer data)
{
	gtk_sess *sess = (gtk_sess *) data;
	gtk_dl_sess *dlsess;
	
	dlsess = gtk_clist_get_row_data(GTK_CLIST(sess->transfer_list), sess->transfer_current_sel);

	if(dlsess != NULL)
	{
		gdk_input_remove(dlsess->socketid);
		close(dlsess->socketfd);
		fclose(dlsess->stream);
		g_free(dlsess);
		gtk_clist_set_row_data(GTK_CLIST(sess->transfer_list), sess->transfer_current_sel, NULL);
		gtk_clist_set_text(GTK_CLIST(sess->transfer_list), sess->transfer_current_sel, 5, TRANSFER_STATUS_CANCEL);
	}
}

void clear_transfer(GtkWidget *widget, gpointer data)
{
	gtk_sess *sess = (gtk_sess *) data;
	
	cancel_download(NULL, data);
	gtk_clist_remove(GTK_CLIST(sess->transfer_list), sess->transfer_current_sel);
	sess->transfer_current_sel = -1;
	gtk_window_set_title(GTK_WINDOW(sess->window), TITLEBAR_MAIN);
}

void selection_unselected(GtkCList *clist, gint row, gint column, GdkEventButton *event, gpointer user_data)
{
	gtk_sess *sess = (gtk_sess *)user_data;
	sess->current_sel = g_list_remove(sess->current_sel, GINT_TO_POINTER(row));
}

void selection_changed(GtkWidget *widget, gint row, gint column, GdkEventButton *event, gpointer data)
{
	gtk_sess *sess = (gtk_sess *)data;
	sess->current_sel = g_list_append(sess->current_sel, GINT_TO_POINTER(row));
}

void transfer_selection_changed(GtkWidget *widget, gint row, gint column, GdkEventButton *event, gpointer data)
{
	gtk_sess *sess = (gtk_sess *)data;
	sess->transfer_current_sel = row;
}

void transfer_selection_unselected(GtkCList *clist, gint row, gint column, GdkEventButton *event, gpointer user_data)
{
	gtk_sess *sess = (gtk_sess *)user_data;
	sess->transfer_current_sel = -1;
	gtk_window_set_title(GTK_WINDOW(sess->window), TITLEBAR_MAIN);
}

gchar *make_file(gchar *in)
{
	gint i;
	gint length = strlen(in);

	for(i=(length-1); i>0; i--)
		if(in[i] == '\\') break;
	i++;		
	return (&in[i]);
}

void get_file_ok(gtk_sess *sess, gchar *instr)
{
	gint row = 0;
	gchar *filename = NULL;
	gchar *tspeed;
	gchar size[10];
	gchar *fsize;
	GdkPixmap *pixm;
	GdkBitmap *bitm;
	gint spacing;
	gint speed;
	gtk_dl_sess *dlsess;
	gchar *seplist[6];
	
	/* 0=xxx 1=IP 2=Port 3=Filename 4=Database 5=Speed */
	
	seperate_string(instr, (gchar **)seplist, 6);
	speed = atoi(seplist[5]);
	tspeed = get_speed(speed);	

	dlsess = (gtk_dl_sess *)g_malloc(sizeof(gtk_dl_sess));
	memset(dlsess, 0, sizeof(gtk_dl_sess));
	dlsess->sess = sess;
	dlsess->lastpercent = -1;

	while((gtk_clist_get_pixtext(GTK_CLIST(dlsess->sess->transfer_list), row, 0, &filename, (guint8*)&spacing, &pixm, &bitm) != 0) &&
		(strcasecmp(filename, make_file(seplist[3])) != 0))
	{
		row++;
		filename = NULL;
	}
	if(filename == NULL) return;
	
	if(connect_to_fileserver(dlsess, seplist[1], seplist[2], seplist[3], row))
	{
		sprintf(size, "%d", (gint)dlsess->size);
		fsize = make_text_size(dlsess->size);
		if(dlsess->sess->SHOW_BYTE)	gtk_clist_set_text(GTK_CLIST(dlsess->sess->transfer_list), row, 1, fsize);
		else gtk_clist_set_text(GTK_CLIST(dlsess->sess->transfer_list), row, 1, size);
		gtk_clist_set_text(GTK_CLIST(dlsess->sess->transfer_list), row, 4, tspeed);
		gtk_clist_set_row_data(GTK_CLIST(dlsess->sess->transfer_list), row, dlsess);
		dlsess->start_time = time(NULL);
		g_free(fsize);
	}
	else
	{
		close(dlsess->socketfd);
		g_free(dlsess);
	}
}

void file_to_read(gpointer data, gint source, GdkInputCondition condition)
{
	gtk_dl_sess *dlsess = (gtk_dl_sess *)data;
	gchar buffer[2048];
	gchar percent_done[15];
	gint length;
	gint row;
	gint percent;
	gchar *msg;
	gdouble dlrate;
	gboolean DONE = FALSE;
	
	row = gtk_clist_find_row_from_data(GTK_CLIST(dlsess->sess->transfer_list), dlsess);
	
	if((length = recv(dlsess->socketfd, buffer, 2048, 0)) < 1)
	{
		if((dlsess->current + length) < dlsess->size)
		{
			perror("recv");
			gtk_clist_set_text(GTK_CLIST(dlsess->sess->transfer_list), row, 5, TRANSFER_STATUS_ERROR);
			gtk_clist_set_row_data(GTK_CLIST(dlsess->sess->transfer_list), row, NULL);
			gdk_input_remove(dlsess->socketid);
			close(dlsess->socketfd);
			fclose(dlsess->stream);
			g_free(dlsess);
			DONE = TRUE;
		}
	}

	if(dlsess->sess->transfer_current_sel == -1) dlsess->lastpercent = -2;
	
	if(!DONE)
	{
		fwrite(buffer, 1, length, dlsess->stream);
		dlsess->current += length;
		if(dlsess->current >= dlsess->size)
		{
			if(dlsess->sess->transfer_current_sel == row)
			{
				percent = 100;
				msg = (gchar *)g_malloc(strlen(TITLEBAR_STATUS) + 4);
				sprintf(msg, TITLEBAR_STATUS, percent);				
				gtk_window_set_title(GTK_WINDOW(dlsess->sess->window), msg);
				g_free(msg);
			}
			gtk_clist_set_text(GTK_CLIST(dlsess->sess->transfer_list), row, 2, "100%");
			gtk_clist_set_text(GTK_CLIST(dlsess->sess->transfer_list), row, 5, TRANSFER_STATUS_FINISH);
			gtk_clist_set_row_data(GTK_CLIST(dlsess->sess->transfer_list), row, NULL);
			gdk_input_remove(dlsess->socketid);
			close(dlsess->socketfd);
			fclose(dlsess->stream);
			g_free(dlsess);
		} 
		else
		{
			dlrate = difftime(time(NULL), dlsess->start_time);
			if(dlrate != 0) dlrate = ((gdouble)dlsess->current/1024)/dlrate;
			percent = (gint)((dlsess->current*100)/dlsess->size);
			if((percent != dlsess->lastpercent) && (dlsess->sess->transfer_current_sel == row))
			{
				msg = (gchar *)g_malloc(strlen(TITLEBAR_STATUS) + 4);
				sprintf(msg, TITLEBAR_STATUS, percent);				
				gtk_window_set_title(GTK_WINDOW(dlsess->sess->window), msg);
				g_free(msg);
				dlsess->lastpercent = percent;
			}
			sprintf(percent_done, "%d%% %.2f KB/s", percent, dlrate);
			gtk_clist_set_text(GTK_CLIST(dlsess->sess->transfer_list), row, 2, percent_done);
		}
	}
}

gint connect_to_fileserver(gtk_dl_sess *dlsess, gchar *ip, gchar *port, gchar *file, gint row)
{
	struct sockaddr_in dest_addr;
	unsigned short int sport;
	gint sendt = 0;
	gint length;
	gchar *msg = "GET";
	gchar buffer[20];
	gchar *command;
	gboolean LOCK;
	gint old;
	fd_set readfds;
	time_t start_time;
   struct timeval tv;

	gchar temp[100];

	sport = atoi(port);
	
	dest_addr.sin_family = AF_INET;
	dest_addr.sin_port = htons(sport);
	dest_addr.sin_addr.s_addr = (unsigned long int)strtoul(ip, NULL, 10);
	memset(dest_addr.sin_zero, 0, 8);
	
	sprintf(temp, "\nIP: %s PORT: %s\n", ip, port);
	g_print(temp);
	
	if((dlsess->socketfd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
	{
		perror("socket");
  	  	gtk_statusbar_push(GTK_STATUSBAR(dlsess->sess->statusbar), 1, STATUSBAR_ERROR_TRANSFER);
  	  	gtk_clist_set_text(GTK_CLIST(dlsess->sess->transfer_list), row, 5, TRANSFER_STATUS_ERROR);
		return FALSE;
	}
	
	gtk_clist_set_text(GTK_CLIST(dlsess->sess->transfer_list), row, 5, TRANSFER_STATUS_CONNECTING);
	
	start_time = time(NULL);
	
	old = fcntl(dlsess->socketfd, F_GETFL);
  	fcntl(dlsess->socketfd, F_SETFL, O_NONBLOCK | old);

  	while(connect(dlsess->socketfd, (struct sockaddr *) &dest_addr, sizeof(struct sockaddr)) != 0)
  	{
		while(gtk_events_pending())
			gtk_main_iteration();
		if(difftime(time(NULL), start_time) > dlsess->sess->timeout)
		{
			gtk_clist_set_text(GTK_CLIST(dlsess->sess->transfer_list), row, 5, TRANSFER_STATUS_TIMEOUT);
			return FALSE;
		}
  	}
	
	fcntl(dlsess->socketfd, F_SETFL, old);
	
	gtk_clist_set_text(GTK_CLIST(dlsess->sess->transfer_list), row, 5, TRANSFER_STATUS_GET);
   
	if((send(dlsess->socketfd, msg, 3, 0)) == -1)
	{
		perror("send");
  	  	gtk_statusbar_push(GTK_STATUSBAR(dlsess->sess->statusbar), 1, STATUSBAR_ERROR_GET);
  	  	gtk_clist_set_text(GTK_CLIST(dlsess->sess->transfer_list), row, 5, TRANSFER_STATUS_ERROR);
		return FALSE;
	}
	
	gtk_clist_set_text(GTK_CLIST(dlsess->sess->transfer_list), row, 5, TRANSFER_STATUS_OK);

	LOCK = TRUE;	
	while(LOCK)
	{
		tv.tv_sec = 0;
		tv.tv_usec = 500000;
		
	  	FD_ZERO(&readfds);
	   FD_SET(dlsess->socketfd, &readfds);
		
		if((select(dlsess->socketfd+1, &readfds, NULL, NULL, &tv)) == -1)
		{
			perror("select");
			gtk_clist_set_text(GTK_CLIST(dlsess->sess->transfer_list), row, 5, TRANSFER_STATUS_ERROR);
			return FALSE;
		}
		if(FD_ISSET(dlsess->socketfd, &readfds)) 
		{		
			recv(dlsess->socketfd, buffer, 6, 0);
			LOCK = FALSE;
		}
		else
		{
			while(gtk_events_pending())
				gtk_main_iteration();
				
			if(difftime(time(NULL), start_time) > dlsess->sess->timeout)
			{
				gtk_clist_set_text(GTK_CLIST(dlsess->sess->transfer_list), row, 5, TRANSFER_STATUS_TIMEOUT);
				return FALSE;
			}
		}
	}
	
	command = (gchar *)g_malloc(strlen((gchar *)dlsess->sess->username) + strlen(file) + 6);
	sprintf(command, "%s \"%s\" 0", (gchar *)dlsess->sess->username, file);
	length = strlen(command);

	gtk_clist_set_text(GTK_CLIST(dlsess->sess->transfer_list), row, 5, TRANSFER_STATUS_FILE);

	FD_ZERO(&readfds);
   FD_SET(dlsess->socketfd, &readfds);
   
	if((sendt += send(dlsess->socketfd, command, length, 0)) == -1)
	{
		perror("send");
		g_free(command);
  		gtk_statusbar_push(GTK_STATUSBAR(dlsess->sess->statusbar), 1, STATUSBAR_ERROR_FILENAME);
  		gtk_clist_set_text(GTK_CLIST(dlsess->sess->transfer_list), row, 5, TRANSFER_STATUS_ERROR);
		return FALSE;
	}
  
	g_free(command);

	gtk_clist_set_text(GTK_CLIST(dlsess->sess->transfer_list), row, 5, TRANSFER_STATUS_SIZE);

	LOCK = TRUE;	
	while(LOCK)
	{
		tv.tv_sec = 0;
		tv.tv_usec = 500000;
		
	  	FD_ZERO(&readfds);
	   FD_SET(dlsess->socketfd, &readfds);

		if(select(dlsess->socketfd+1, &readfds, NULL, NULL, &tv) == -1)
		{
			perror("select");
			gtk_clist_set_text(GTK_CLIST(dlsess->sess->transfer_list), row, 5, TRANSFER_STATUS_ERROR);
			return FALSE;
		}
		
		if(FD_ISSET(dlsess->socketfd, &readfds)) 
		{		
			length = recv(dlsess->socketfd, buffer, 20, 0);
			LOCK = FALSE;
		}
		else
		{
			while(gtk_events_pending())
				gtk_main_iteration();
				
			if(difftime(time(NULL), start_time) > dlsess->sess->timeout)
			{
				gtk_clist_set_text(GTK_CLIST(dlsess->sess->transfer_list), row, 5, TRANSFER_STATUS_TIMEOUT);
				return FALSE;
			}
		}
	}
	
	buffer[length] = '\0';

	if(strcmp(buffer, "INVALID REQUEST") == 0) 
	{
		gtk_statusbar_push(GTK_STATUSBAR(dlsess->sess->statusbar), 1, STATUSBAR_ERROR_INVALID_REQUEST);
		gtk_clist_set_text(GTK_CLIST(dlsess->sess->transfer_list), row, 5, TRANSFER_STATUS_ERROR);
		return FALSE;
	}
	
	if(strcmp(buffer, "FILE NOT SHARED") == 0) 
	{
		gtk_statusbar_push(GTK_STATUSBAR(dlsess->sess->statusbar), 1, STATUSBAR_ERROR_NOT_SHARED);
		gtk_clist_set_text(GTK_CLIST(dlsess->sess->transfer_list), row, 5, TRANSFER_STATUS_ERROR);
		return FALSE;
	}
	
	if(strcmp(buffer, "FILE NOT FOUND") == 0) 
	{
		gtk_statusbar_push(GTK_STATUSBAR(dlsess->sess->statusbar), 1, STATUSBAR_ERROR_NOT_FOUND);
		gtk_clist_set_text(GTK_CLIST(dlsess->sess->transfer_list), row, 5, TRANSFER_STATUS_ERROR);
		return FALSE;
	}
	
	command = (gchar *)g_malloc(strlen(file) + strlen(dlsess->sess->directory) + 2);
	
	strcpy(command, dlsess->sess->directory);
	if(dlsess->sess->directory[strlen(dlsess->sess->directory)-1] != '/')
		strcat(command, "/");
	strcat(command, make_file(file));

	if((dlsess->stream = fopen(command, "wb")) == NULL)
	{
		gtk_statusbar_push(GTK_STATUSBAR(dlsess->sess->statusbar), 1, STATUSBAR_ERROR_FILE);
		g_print(command);
		perror("fopen");
		g_free(command);
		gtk_clist_set_text(GTK_CLIST(dlsess->sess->transfer_list), row, 5, TRANSFER_STATUS_ERROR);
		return FALSE;
	}
	
	g_free(command);

	dlsess->size = atol(buffer);
	dlsess->socketid = gdk_input_add(dlsess->socketfd, GDK_INPUT_READ, GTK_SIGNAL_FUNC(file_to_read), dlsess);	

	gtk_clist_set_text(GTK_CLIST(dlsess->sess->transfer_list), row, 5, TRANSFER_STATUS_DOWNLOAD);

	return TRUE;
}

